"""
Created on 2025-02-02

@author: wf
"""


class VarNameTracker:
    """
    Tracks variable names and ensures uniqueness by appending a suffix (_2, _3, etc.).
    """

    def __init__(self):
        self.var_name_count = {}

    def get_unique_name(self, base_name):
        """
        Get a unique variable name by appending a suffix if needed.
        """
        if base_name in self.var_name_count:
            self.var_name_count[base_name] += 1
            return f"{base_name}_{self.var_name_count[base_name]}"
        else:
            self.var_name_count[base_name] = 1
            return base_name


class QueryGen:
    """
    Generator for SPARQL queries based on property count query results.
    """

    def __init__(self, prefixes, debug: bool = False):
        """
        Initialize the QueryGen with a dictionary of prefixes.
        """
        self.prefixes = prefixes
        self.debug = debug

    def sanitize_variable_name(self, prop):
        """
        Convert a prefixed prop into a valid SPARQL variable name.
        """
        parts = prop.split(":")
        var_name = ""
        if parts:
            var_name = parts[-1]
            for invalid in ["-", ":", "#"]:
                var_name = var_name.replace(invalid, "_")
        var_name = f"{var_name}"
        return var_name

    def get_prefixed_property(self, prop: str) -> str:
        """
        Convert a full URI into a prefixed SPARQL property if a matching prefix is found.

        Args:
            prop (str): The full URI of the property.

        Returns:
            str: The prefixed property if a matching prefix is found, otherwise the original URI in angle brackets.
        """
        longest_match = None
        for prefix, uri in self.prefixes.items():
            if prop.startswith(uri) and (
                longest_match is None or len(uri) > len(longest_match)
            ):
                longest_match = uri
                best_prefix = prefix

        if longest_match:
            return prop.replace(longest_match, f"{best_prefix}:")

        return f"<{prop}>"

    def gen(
        self,
        lod,
        main_var: str,
        main_value: str,
        max_cardinality: int = None,
        first_x: int = None,
        comment_out: bool = False,
    ):
        """
        Generates a SPARQL query dynamically based on the provided list of dictionaries (lod).

        Args:
            lod (list[dict]): List of dictionaries containing property data.
            main_var (str): The main variable in the SPARQL query.
            main_value (str): The value assigned to the main variable.
            max_cardinality (int, optional): Maximum allowed cardinality for properties. Defaults to a large number.
            first_x (int, optional): Number of properties to include before commenting out the rest. Defaults to a large number.
            comment_out (bool, optional): If True, comments out properties that are filter by the first_x or max cardinality condition

        Returns:
            str: The generated SPARQL query as a string.
        """
        if first_x is None:
            first_x = 10**9  # a billion properties? should not happen
        if max_cardinality is None:
            max_cardinality = 10**16  # how much energy for the RAM to keep this?
        sparql_query = "# generated Query"
        properties = {}
        tracker = VarNameTracker()
        for i, record in enumerate(lod):
            prop = record["p"]
            # do we have an injected wikidata_property?
            wikidata_property = record.get("wikidata_property")
            prefixed_prop = self.get_prefixed_property(prop)
            base_var_name = self.sanitize_variable_name(prefixed_prop)
            var_name = tracker.get_unique_name(base_var_name)
            card = int(record.get("count", 1))
            comment = "" if i < first_x and card <= max_cardinality else "#"
            properties[var_name] = (
                prop,
                prefixed_prop,
                card,
                comment,
                wikidata_property,
            )
        for key, value in self.prefixes.items():
            sparql_query += f"\nPREFIX {key}: <{value}>"

        sparql_query += f"\nSELECT ?{main_var}"

        for i, (var_name, (prop, prefixed_prop, card, comment, wdp)) in enumerate(
            properties.items()
        ):
            if not (comment_out and comment):
                sparql_query += f"\n  {comment}?{var_name}"

        sparql_query += "\nWHERE {\n"

        sparql_query += f"  VALUES (?{main_var}) {{ ({main_value}) }}\n"
        sparql_query += "  OPTIONAL {\n"

        for i, (var_name, (prop, prefixed_prop, card, comment, wdp)) in enumerate(
            properties.items()
        ):
            if not (comment_out and comment):
                sparql_query += f"    # {prop}\n"
                if (
                    wdp
                ):  # label but not description to avoid url encoding and other issues like param length overrun
                    sparql_query += f"    # {wdp.plabel}\n"
                sparql_query += (
                    f"    {comment}?{main_var} {prefixed_prop} ?{var_name} .\n"
                )

        sparql_query += "  }\n"  # Closing Optional clause
        sparql_query += "}"  # Closing WHERE clause

        return sparql_query



"""
Created on 2025-01-17

@author: wf
"""

import argparse
import re
from typing import Optional

import gpxpy
import requests
from ngwidgets.leaflet_map import LeafletMap

from velorail.tour import Leg, LegStyle, LegStyles, Loc, Tour


class GPXViewer:
    """
    display a given gpx file
    """

    samples = {
        "Mountain bike loop at Middlesex Fells reservation.": "https://www.topografix.com/fells_loop.gpx",
        "Via Verde de Arditurri": "https://www.bahntrassenradwege.de/images/Spanien/Baskenland/V-v-de-arditurri/Arditurri2.gpx",
        "Vía Verde del Fc Vasco Navarro": "https://ebike.bitplan.com/images/ebike/e/ec/VV-del-FC-Vasco-Navarro_fdmi1644.gpx",
    }
    default_center = [51.4934, 0.0098]  # greenwich
    default_zoom = 11

    @classmethod
    def from_url(cls, gpx_url) -> "GPXViewer":
        viewer = cls()
        viewer.load_gpx(gpx_url)
        return viewer

    def __init__(self, args: argparse.Namespace = None):
        """
        constructor

        Args:
            args(argparse.Namespace): command line arguments
        """
        self.args = args
        self.tour = None
        self.leg_styles = LegStyles.default()
        self.set_center()
        if args:
            self.debug = args.debug
            self.token = args.token
            self.center = args.center
            self.zoom = args.zoom
            if self.args.gpx:
                self.load_gpx(self.args.gpx)
        else:
            self.zoom = GPXViewer.default_zoom
            self.center = GPXViewer.default_center


    def load_gpx(self, gpx_url: str):
        """
        Load the given GPX file from URL and parse it.
        """
        try:
            response = requests.get(gpx_url, timeout=10)
            response.raise_for_status()
            self.gpx = gpxpy.parse(response.text)
            self.get_points(self.gpx)
            return self.gpx
        except (requests.RequestException, gpxpy.gpx.GPXXMLSyntaxException) as e:
            raise RuntimeError(f"Failed to load GPX from {gpx_url}: {e}")


    def set_center(self):
        """
        Calculate and set the center and bounding box based on tour legs
        """

        if self.tour and self.tour.legs:
            points = []
            for leg in self.tour.legs:
                points.append(leg.start.coordinates)
                points.append(leg.end.coordinates)
            # Wrong order: lats, lons need to be extracted after zipping
            lats = [p[0] for p in points]
            lons = [p[1] for p in points]
            self.bounding_box = (min(lats), max(lats), min(lons), max(lons))
            self.center = ((min(lats) + max(lats)) / 2, (min(lons) + max(lons)) / 2)
        else:
            self.center = self.default_center
            self.bounding_box = None
        return self.center

    def add_leg(
        self,
        start_point,
        end_point,
        leg_type: str,
        add_end_point: bool = False,
        url: Optional[str] = None,
    ):
        """
        Add a leg to the tour

        Args:
            start_point: GPX point for start of leg
            end_point: GPX point for end of leg
            leg_type: Type of leg (e.g., "bike", "train", "car")
            add_end_point: Whether to add the end point (True for last leg)
            url: Optional URL associated with the leg
        """
        if self.tour is None:
            self.tour = Tour(name="GPX Tour")

        # Create locations
        start_loc = Loc(
            id=str(len(self.tour.legs)),
            name=start_point.name if hasattr(start_point, "name") else None,
            coordinates=(start_point.latitude, start_point.longitude),
        )
        end_loc = Loc(
            id=str(len(self.tour.legs) + 1),
            name=end_point.name if hasattr(end_point, "name") else None,
            coordinates=(end_point.latitude, end_point.longitude),
        )

        # Create and add leg
        leg = Leg(leg_type=leg_type, start=start_loc, end=end_loc, url=url)
        self.tour.legs.append(leg)

    def get_points(self, gpx, way_points_fallback: bool = False):
        """
        Extract waypoints and legs from the GPX object and create a tour
        """
        self.tour = Tour(name="GPX Tour")

        # Process routes
        for route in gpx.routes:
            for i in range(len(route.points) - 1):
                url = route.link.href if route.link else None
                is_last = i == len(route.points) - 2
                self.add_leg(
                    route.points[i],
                    route.points[i + 1],
                    "bike",  # Default to bike for routes
                    add_end_point=is_last,
                    url=url,
                )

        # Process tracks
        for track in gpx.tracks:
            for segment in track.segments:
                for i in range(len(segment.points) - 1):
                    is_last = i == len(segment.points) - 2
                    self.add_leg(
                        segment.points[i],
                        segment.points[i + 1],
                        "bike",  # Default to bike for tracks
                        add_end_point=is_last,
                    )

        prev_loc = None
        # Handle waypoints if no legs were created and fallback is active
        if not self.tour.legs and gpx.waypoints and way_points_fallback:
            for i, waypoint in enumerate(gpx.waypoints):
                loc = Loc(
                    id=str(i),
                    name=waypoint.name,
                    coordinates=(waypoint.latitude, waypoint.longitude),
                    notes=waypoint.description,
                )
                if i > 0:
                    leg = Leg(
                        leg_type="bike",
                        start=prev_loc,
                        end=loc,
                        url=waypoint.link.href if waypoint.link else None,
                    )
                    self.tour.legs.append(leg)
                prev_loc = loc

    def parse_lines(self, lines: str):
        """
        Parse the 'lines' parameter into route segments.

        Args:
            lines (str): The input string containing routes in the format:
                "By bike: 51.243931° N, 6.520022° E, 51.269222° N, 6.625467° E:"

        Returns:
            Tour: Created tour from the parsed coordinates
        """
        coordinate_pattern = r"(\d+\.\d+)° ([NS]), (\d+\.\d+)° ([EW])"
        route_segments = lines.split(":")
        self.tour = Tour(name="Parsed Tour")

        for segment in route_segments:
            segment = segment.strip()

            # Extract leg type (e.g., "bike", "train")
            leg_type_match = re.match(r"By (\w+)→", segment)
            if leg_type_match:
                leg_type = leg_type_match.group(1).lower()
                pass
            else:
                leg_type = "bike"

            points = re.findall(coordinate_pattern, segment)
            if points:
                prev_loc = None
                for i, (lat, ns, lon, ew) in enumerate(points):
                    lat_val = float(lat) * (-1 if ns == "S" else 1)
                    lon_val = float(lon) * (-1 if ew == "W" else 1)

                    loc = Loc(
                        id=str(len(self.tour.legs) + i),
                        name=None,
                        coordinates=(lat_val, lon_val),
                    )

                    if prev_loc:
                        leg = Leg(leg_type=leg_type, start=prev_loc, end=loc)
                        self.tour.legs.append(leg)
                    prev_loc = loc

        if not self.tour.legs:
            raise ValueError("No valid routes found in the input lines.")

        return self.tour

    def parse_lines_and_show(self, lines: str, zoom: int = None):
        """
        Parse lines and display them on the map
        """
        self.parse_lines(lines)
        self.show(zoom=zoom)

    def show(self, zoom: int = None, center=None):
        """
        Show tour with styled paths
        """
        if zoom is None:
            zoom = self.zoom
        if center is None:
            center = self.set_center()

        self.map = LeafletMap(center=center, zoom=zoom)
        if self.tour:
            self.map.draw_tour(self.tour, self.leg_styles)



"""
Knowledge Graph walker implementation
Created on 2025-02-13
@author: wf
"""

from typing import Any, Dict, List, Optional

from velorail.explore import Explorer, TriplePos
from velorail.querygen import QueryGen


class KGWalker:
    """
    Knowledge Graph walker that combines exploration, 
    query generation and result viewing
    """

    def __init__(self, endpoint_name: str, debug: bool = False):
        """
        Initialize the KG walker

        Args:
            endpoint_name(str): the name of the endpoint to query
            debug(bool): if True show debug information
        """
        self.endpoint_name = endpoint_name
        self.debug = debug
        self.explorer = Explorer(endpoint_name)
        self.prefixes = self.explorer.endpoint_prefixes.get(endpoint_name)
        self.query_gen = QueryGen(prefixes=self.prefixes, debug=self.debug)

    def get_gen_lod(self, query_lod: list, selected_props: list) -> list:
        """
        Get the table of properties to be generated

        Args:
            query_lod (list): List of dictionaries containing query results
            selected_props (list): List of property IDs to select

        Returns:
            list: Generated list of dictionaries containing selected properties
        """
        seen_props = set()
        gen_lod = []
        # Collect selected properties from query results
        for record in query_lod:
            record_copy = record.copy()
            prop_value = record_copy.get("p")
            prop = self.explorer.get_prop(prop_value)
            if (
                prop
                and prop.pid in selected_props
                and prop.pid not in seen_props
                and "/direct/" in prop_value
            ):
                if self.debug:
                    print(
                        f"selecting {prop} with type_name {prop.type_name} from {record}"
                    )
                record_copy["wikidata_property"] = prop
                gen_lod.append(record_copy)
                seen_props.add(prop.pid)
        return gen_lod

    def explore_and_query(
        self, prefix: str, node_id: str, selected_props: List[str]
    ) -> List[Dict[str, Any]]:
        """
        Explore the knowledge graph and generate/execute query from the given start node

        Args:
            prefix (str): the prefix to use e.g. wd
            node_id (str): the id of the node to start from
            selected_props (List[str]): list of properties to select e.g. ["named after", "P10689"]

        Returns:
            List[Dict[str, Any]]: List of raw query results
        """
        start_node = self.explorer.get_node(node_id, prefix)

        # Get exploration query results
        query_lod = self.explorer.explore_node(
            node=start_node, triple_pos=TriplePos.SUBJECT, summary=True
        )

        # Table of selected properties for generation
        gen_lod = self.get_gen_lod(query_lod, selected_props)

        # Generate query with selected properties
        generated_query = self.query_gen.gen(
            lod=gen_lod, main_var="item", main_value=start_node.qualified_name
        )

        if self.debug:
            print(generated_query)

        # Execute query and store results
        query_results = self.explorer.query(
            sparql_query=generated_query, param_dict={}, endpoint=self.endpoint_name
        )
        return query_results

    def create_view(
        self, query_results: List[Dict[str, Any]], depth: int = 1
    ) -> Optional[Dict[str, Any]]:
        """
        Create a view record from query results

        Args:
            query_results (List[Dict[str, Any]]): List of raw query results
            depth (int): Depth for view record generation, defaults to 1

        Returns:
            Optional[Dict[str, Any]]: View record if results exist, None otherwise
        """
        view_record = None
        if len(query_results) >= 1:
            record = query_results[0]
            view_record = self.explorer.get_view_record(record, depth)
        return view_record

    def walk(
        self,
        prefix: str,
        node_id: str,
        selected_props: List[str],
        create_view: bool = True,
        view_depth: int = 1,
    ) -> Optional[Dict[str, Any]]:
        """
        Walk the knowledge graph from the given start node

        Args:
            prefix (str): the prefix to use e.g. wd
            node_id (str): the id of the node to start from
            selected_props (List[str]): list of properties to select e.g. ["named after", "P10689"]
            create_view (bool): if True, generate a view record from results
            view_depth (int): depth for view record generation if create_view is True, defaults to 1

        Returns:
            Optional[Dict[str, Any]]: View record if create_view is True, raw query results otherwise
        """
        # Get query results
        query_results = self.explore_and_query(prefix, node_id, selected_props)

        # Prepare return value based on mode and results
        result = None
        if query_results:
            if create_view:
                result = self.create_view(query_results, view_depth)
            else:
                result = query_results

        return result



"""
Created on 2025-02-04

@author: wf
"""

import json
import os
from argparse import Namespace

from ngwidgets.basetest import Basetest

from velorail.locfind import NPQ_Handler
from velorail.osm2wiki import Osm2WikiConverter


class TestOsm2wiki(Basetest):
    """
    test  osm2wiki script
    """

    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.tmp_path = "/tmp"
        self.query_handler = NPQ_Handler(
            yaml_file="osmplanet_explore.yaml", debug=self.debug
        )
        self.prefixes = {
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "geo": "http://www.opengis.net/ont/geosparql#",
            "geof": "http://www.opengis.net/def/function/geosparql/",
            "ogc": "http://www.opengis.net/rdf#",
            "osmkey": "https://www.openstreetmap.org/wiki/Key:",
            "osm2rdfmember": "https://osm2rdf.cs.uni-freiburg.de/rdf/member#",
            "osmrel": "https://www.openstreetmap.org/relation/",
            "osm2rdf": "https://osm2rdf.cs.uni-freiburg.de/rdf/",
            "osm2rdf_geom": "https://osm2rdf.cs.uni-freiburg.de/rdf/geom#",
            "meta": "https://www.openstreetmap.org/meta/",
            "xsd": "http://www.w3.org/2001/XMLSchema#",
        }

    def testExplore(self):
        """
        test queries to explore correct OSM Planet SPAQRL queries
        """
        for query_name in ["Relation1", "RelationExplore"]:
            lod = self.query_handler.query_by_name(
                query_name=query_name,
                param_dict={"relid": "10492086"},
                endpoint="osm-qlever",
            )
            if self.debug:
                print(f"Query: {query_name}:")
                print(json.dumps(lod, indent=2))

    def testCompressNodes(self):
        """
        Test node compression with different scenarios.
        """
        # Create a straight line of points spaced 500m apart
        points = [
            {"loc": f"POINT(2.0 {42.0 + (i * 0.004491)})", "node": f"node/{i}"}
            for i in range(10)
        ]

        args = Namespace(debug=True, tmp="/tmp")
        converter = Osm2WikiConverter(args=args)

        # Test cases: (input_points, min_distance_m, max_nodes, expected_count)
        for i, (input_points, min_distance_m, expected_count) in enumerate(
            [
                (points, 1000, 4),  # 1km spacing keeps every other point + last point
                (points, 2000, 2),  # 2km spacing keeps every fourth point + last point
                (
                    points,
                    100,
                    10,
                ),  # max_nodes limits to 4 points (last point may not be included)
                ([], 1000, 0),  # empty input
                ([points[0]], 1000, 1),  # single point
                (points[:2], 1000, 2),  # two points
            ]
        ):

            converter.set_wkts(input_points)
            compressed = converter.compress_nodes(
                input_points, min_distance_m=min_distance_m
            )
            self.assertEqual(len(compressed), expected_count, f"case {i}")

    def testOsmRelConverter(self):
        """
        test the converter
        """
        debug=self.debug
        #debug=True
        for osm_item, role, transport, loc_type, expected_nodes, min_node_distance in [
            ("relation/18343947", "member", "train", "train station",  53, 6000),
            ("relation/3421095" , "member", "bike",  "bike-waypoint",  29, 2000),
            ("relation/1713826" , "member", "bike",  "bike-waypoint",  10, 2500),
            ("relation/10492086", "stop"  , "train", "train station",  21, 8000),
        ]:
            args = Namespace(
                debug=self.debug,
                tmp=self.tmp_path,
                endpoint_name="osm-qlever",
                zoom=8,
                min_lat=42.0,
                max_lat=44.0,
                min_lon=-9.0,
                max_lon=4.0,
                transport=transport,
                loc_type=loc_type,
                role=role,
                country="Spanien",
                category="Spain2025",
                osm_items=[osm_item],
                queriesPath=None,
                queryName="ItemNodesGeo",
                min_node_distance=min_node_distance,
            )
            # Create converter instance
            converter = Osm2WikiConverter(args=args)
            converter.test = True

            # Process the relations
            lod = converter.process_osm_items(args.osm_items)
            node_count=len(lod)
            if debug:
                print(f"{loc_type}:{osm_item} {node_count} nodes")
                print(json.dumps(lod, indent=2, default=str))
                print(f"{loc_type}:{osm_item} {node_count} nodes expected: {expected_nodes}")

            self.assertTrue(node_count >= expected_nodes,node_count)
            self.assertTrue(os.path.exists(converter.wiki_file))



"""
Created on 2025-02-15

@author: wf
"""

from fastapi.responses import RedirectResponse
from ngwidgets.login import Login
from ngwidgets.webserver import WebSolution, NiceGuiWebserver
from nicegui import Client, ui
from wikibot3rd.sso import User
from wikibot3rd.sso_users import Sso_Users

class SsoSolution(WebSolution):
    """
    Encapsulates SSO authentication setup and user handling
    """
    def __init__(self, webserver: NiceGuiWebserver, client: Client = None, credentials_path=None):
        """
        Initialize SSO authentication

        Args:
            webserver: The webserver instance to set up SSO for
            client: The client instance (can be None for non-client specific solutions)
            credentials_path: Optional path to the SSO credentials file
        """
        super().__init__(webserver, client)  # Call parent constructor with both required parameters
        self.users = Sso_Users(webserver.config.short_name, credentials_path=credentials_path)

        self.login = Login(webserver, self.users)

    def get_user(self) -> User:
        """
        Get the current authenticated user

        Returns:
            User: The current user object or None if not authenticated
        """
        user = None
        username = self.login.get_username()
        if username and self.users.is_available:
            user = self.users.sso.get_user(username)
        return user

    def get_user_display_name(self) -> str:
        """
        Get the display name for the current user with admin indicator

        Returns:
            str: Username with optional admin indicator
        """
        username = self.login.get_username()
        if username is None:
            username = "?"
        user = self.get_user()
        admin_flag = "🔑" if user and user.is_admin else ""
        user_display = f"{username}{admin_flag}"
        return user_display

    async def logout(self):
        """Handle user logout"""
        await self.login.logout()

    def as_html(self) -> str:
        """
        Get the user details as HTML markup

        Returns:
            str: HTML markup for the user details or error message
        """
        user = self.get_user()
        html_markup = "<h1>User Details</h1>"
        if not user:
            html_markup += "<p>No user logged in</p>"
        else:
            html_markup += f"""
            <p><strong>ID:</strong> {user.id}</p>
            <p><strong>Name:</strong> {user.name}</p>
            <p><strong>Real Name:</strong> {user.real_name}</p>
            <p><strong>Email:</strong> {user.email}</p>
            <p><strong>Edit Count:</strong> {user.editcount}</p>
            """
        return html_markup

    async def show_login(self):
        """Show the login page"""
        await self.login.login(self)

    async def show_user_details(self):
        """Show the user details page"""
        def show():
            self.logout_button = ui.button(
                "logout", icon="logout",
                on_click=self.logout
            )
            ui.html(self.as_html())
        await self.setup_content_div(show)

    def configure_menu(self):
        """Configure the user menu"""
        display_name = self.get_user_display_name()
        self.link_button(display_name, "/user", "person")
        # make sure the link exists
        self.register_pages()

    def register_pages(self):
        """Register the SSO-related pages"""
        @ui.page("/user")
        async def show_user(client: Client):
            if not  self.login.authenticated():
                return RedirectResponse("/login")
            return await self.webserver.execute_action(
                client,
                solution_class=SsoSolution,
                wanted_action=SsoSolution.show_user_details
            )

        @ui.page("/login")
        async def login(client: Client):
            return await self.webserver.execute_action(
                client,
                solution_class=SsoSolution,
                wanted_action=SsoSolution.show_login
            )


"""
Created on 2025-06-02

@author: wf
"""

from ngwidgets.lod_grid import GridConfig, ListOfDictsGrid
from ngwidgets.webserver import WebSolution
from nicegui import ui

from velorail.explore import Explorer, TriplePos
from velorail.querygen import QueryGen
from ngwidgets.task_runner import TaskRunner


class ExplorerView:
    """SPARQL Explorer nicegui component"""

    def __init__(
        self,
        solution: WebSolution,
        prefix: str = "osm:relation",
        endpoint_name: str = "osm-qlever",
        summary: bool = False,
    ):
        """Initialize the explorer view"""
        self.solution = solution
        self.args = self.solution.args
        self.prefix = prefix
        self.endpoint_name = endpoint_name
        self.summary = summary
        self.explorer = Explorer(endpoint_name)
        self.result_row = None
        self.lod_grid = None
        self.node_id = None
        self.task_runner=TaskRunner()


    async def on_generate_query(self):
        """Handle query generation from selected rows"""
        try:
            lod_index=self.lod_grid.get_index(lenient=self.lod_grid.config.lenient,lod=self.lod)
            lod = await self.lod_grid.get_selected_lod(lod_index=lod_index)
            if len(lod)==0:
                with self.result_row:
                    ui.notify("Please select at least one row")
                    return
            prefixes = self.explorer.endpoint_prefixes.get(self.endpoint_name)
            query_gen = QueryGen(prefixes, debug=self.args.debug)
            sparql_query = query_gen.gen(
                lod, main_var="item", main_value=f"{self.prefix}:{self.node_id}"
            )
            self.query_code.content = sparql_query
        except Exception as ex:
            self.solution.handle_exception(ex)

    async def on_run_query(self):
        """
        Run the given SPARQL query and display the result.
        """
        try:
            if not self.query_code.content.strip():
                with self.result_row:
                    ui.notify("No query generated. Please generate a query first.")
                return

            sparql_query = self.query_code.content.strip()

            # Execute the query
            lod = self.explorer.query(
                sparql_query=sparql_query, param_dict={}, endpoint=self.endpoint_name
            )

            # Ensure only one record is expected
            if not lod:
                with self.result_row:
                    ui.notify("Query returned no results.")
                return

            if len(lod) > 1:
                with self.result_row:
                    ui.notify(
                        "Unexpected multiple results. Displaying first result only."
                    )

            # Transform result into a single-row format for display
            record = lod[0]  # Expecting a single record
            single_record_lod = [
                {"Property": key, "Value": value} for key, value in record.items()
            ]
            self.update_lod(single_record_lod, with_select=False)

        except Exception as ex:
            self.solution.handle_exception(ex)

    def get_default_query(self) -> str:
        """ """
        if self.node_id is None:
            node_id = "10492086"
        else:
            node_id = self.node_id
        query = (
            """PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX geo: <http://www.opengis.net/ont/geosparql#>
PREFIX osm2rdfmember: <https://osm2rdf.cs.uni-freiburg.de/rdf/member#>
PREFIX osmrel: <https://www.openstreetmap.org/relation/>
PREFIX osmkey: <https://www.openstreetmap.org/wiki/Key:>
PREFIX geof: <http://www.opengis.net/def/function/geosparql/>
SELECT
  ?rel
  ?rel_name
  ?rel_pos
  ?member
  ?node
  ?node_name
  ?node_ref
  ?role
  ?lat
  ?lon
WHERE {
  VALUES (?rel ?role ?minlat ?maxlat ?minlon ?maxlon) {
    (osmrel:"""
            + node_id
            + """ "stop" 42.0 44.0 -9.0 4.0)
  }
  ?rel osmkey:ref ?rel_name .
#  ?rel osmrel:member_ref ?member .
#  ?member osmrel:member_pos ?rel_pos .
#  ?member osmrel:member_role ?role .
#  ?member osmrel:member_id ?node .
#  OPTIONAL { ?node osmkey:name ?node_name}.
#  OPTIONAL { ?node osmkey:ref ?node_ref}.
#  ?node geo:hasGeometry/geo:asWKT ?loc .
#  BIND(geof:latitude(?loc) AS ?lat)
#  BIND(geof:longitude(?loc) AS ?lon)
#  FILTER (?lat > ?minlat && ?lat < ?maxlat && ?lon > ?minlon && ?lon < ?maxlon)
}
#ORDER BY ?rel_pos
"""
        )
        return query

    def update_node_info(self):
        """
        update the node info
        """
        info = f"{self.prefix}:{self.node_id} Endpoint: {self.endpoint_name}"
        with self.header_row:
            self.node_info.content = info

    def setup_ui(self):
        """Setup the basic UI container"""
        with ui.splitter() as splitter:
            with splitter.before:
                self.header_row = ui.row()
                with self.header_row:
                    self.node_info = ui.html(f"")
                    ui.button(
                        "Generate Query", icon="code", on_click=self.on_generate_query
                    )
                    self.result_row = ui.row().classes("w-full")
            with splitter.after:
                ui.button("run", icon="run", on_click=self.on_run_query)
                self.query_code = (
                    ui.code(content=self.get_default_query(), language="SPARQL")
                    .classes("w-full")
                    .props("rows=10")
                )

    def show(self, node_id: str):
        """Show exploration results for a given node ID"""
        self.node_id = node_id
        self.update_node_info()
        self.task_runner.run(self.explore_node_task)

    async def explore_node_task(self):
        """Background task for node exploration"""
        try:
            # Clear and show loading state
            self.result_row.clear()
            with self.result_row:
                ui.spinner("dots")
            self.result_row.update()

            # Get exploration results
            start_node = self.explorer.get_node(self.node_id, self.prefix)
            with self.result_row:
                ui.label(
                    f"Exploring {start_node.qualified_name} on {self.endpoint_name}"
                )

            lod = self.explorer.explore_node(
                node=start_node, triple_pos=TriplePos.SUBJECT, summary=self.summary
            )

            if not lod:
                with self.solution.container:
                    ui.notify("Exploration returned no results")
                return

            self.result_row.clear()
            self.update_lod(lod)

        except Exception as ex:
            self.solution.handle_exception(ex)
            with self.solution.container:
                ui.notify(f"Exploration failed: {str(ex)}")

    def get_view_lod(self, lod: list) -> list:
        """
        Convert records to view format with row numbers and links
        """
        view_lod = []
        for i, record in enumerate(lod):
            index = i + 1
            view_record = self.explorer.get_view_record(record, index)
            view_lod.append(view_record)
        return view_lod

    def update_lod(self, lod: list, with_select: bool = True):
        """Update grid with list of dicts data"""
        self.lod=lod
        view_lod = self.get_view_lod(lod)
        # Configure grid with checkbox selection
        grid_config = GridConfig(
            key_col="#",
            editable=False,
            multiselect=True,
            with_buttons=True,
            button_names=["all", "fit"],
            debug=False,
        )

        # Create or update grid
        if self.lod_grid is None:
            with self.result_row:
                self.lod_grid = ListOfDictsGrid(lod=view_lod, config=grid_config)
        else:
            with self.result_row:
                self.lod_grid.load_lod(view_lod)
                self.lod_grid.update()
                self.lod_grid.sizeColumnsToFit()
        with self.result_row:
            if with_select:
                self.lod_grid.set_checkbox_selection("#")



"""
Created on 2025-02-13

@author: wf
"""

import json

from ngwidgets.basetest import Basetest

from velorail.kgwalker import KGWalker


class TestKGWalker(Basetest):
    """
    test Knowledge Graph Walker
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_kg_walk(self):
        """
        Test exploring Madeira airport looking for specific property labels
        """
        test_cases = [
            ("named after", "P138"),  # entity that inspired the name
            ("OpenStreetMap way ID", "P10689"),  # identifier for OSM way
        ]
        prefix = "wd"
        node_id = "Q639161"  # Madeira Airport

        # get pids for query
        selected_props = [pid for _label, pid in test_cases]

        walker = KGWalker(endpoint_name="wikidata", debug=self.debug)
        view_record = walker.walk(
            prefix=prefix, node_id=node_id, selected_props=selected_props
        )

        # check that we got a result
        msg = "Expected view record but got None"
        self.assertIsNotNone(view_record, msg)

        if self.debug:
            print(json.dumps(view_record, indent=2, default=str))

        # verify each property is found with correct label
        for label, pid in test_cases:
            found = False
            for value in view_record.values():
                value_str = str(value)
                if label in value_str and pid in value_str:
                    found = True
                    break
            msg = f"Property {pid} with label {label} not found in view record"
            if self.debug:
                print(msg)
            # FIXME - enable this test
            # self.assertTrue(found, msg)



"""
Created on 2025-02-01

@author: wf
"""

import sys
from argparse import ArgumentParser

from lodstorage.query import EndpointManager
from ngwidgets.cmd import WebserverCmd

from velorail.gpxviewer import GPXViewer
from velorail.webserver import VeloRailWebServer


class VeloRailCmd(WebserverCmd):
    """
    command line handling for velorail
    """

    def __init__(self):
        """
        constructor
        """
        config = VeloRailWebServer.get_config()
        WebserverCmd.__init__(self, config, VeloRailWebServer)
        pass

    def getArgParser(self, description: str, version_msg) -> ArgumentParser:
        """
        override the default argparser call
        """
        parser = super().getArgParser(description, version_msg)
        parser.add_argument(
            "-en",
            "--endpointName",
            default="wikidata",
            help=f"Name of the endpoint to use for queries. Available by default: {EndpointManager.getEndpointNames(lang='sparql')}",
        )
        parser.add_argument(
            "--lang", type=str, default="en", help="Language for the UI (de or en)"
        )
        parser.add_argument(
            "-rp",
            "--root_path",
            default=VeloRailWebServer.examples_path(),
            help="path to velorail files [default: %(default)s]",
        )
        parser.add_argument("--gpx", required=False, help="URL or path to GPX file")
        parser.add_argument(
            "--token", required=False, help="Authentication token for GPX access"
        )
        parser.add_argument(
            "--zoom",
            type=int,
            default=GPXViewer.default_zoom,
            help="zoom level (default: 11)",
        )
        parser.add_argument(
            "--center",
            nargs=2,
            type=float,
            default=GPXViewer.default_center,
            help="center lat,lon - default: Greenwich",
        )
        return parser


def main(argv: list = None):
    """
    main call
    """
    cmd = VeloRailCmd()
    exit_code = cmd.cmd_main(argv)
    return exit_code


if __name__ == "__main__":
    sys.exit(main())



"""
Created on 2025-02-09

@author: wf
"""

import json

from ngwidgets.basetest import Basetest

from velorail.explore import Explorer, TriplePos
from velorail.npq import NPQ_Handler
from velorail.querygen import QueryGen


class TestQueryGen(Basetest):
    """
    test SPARQL Query Generator for explorer
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.query_handler = NPQ_Handler(
            yaml_file="osmplanet_explore.yaml", debug=self.debug
        )

        self.prefixes = {
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "geo": "http://www.opengis.net/ont/geosparql#",
            "geof": "http://www.opengis.net/def/function/geosparql/",
            "ogc": "http://www.opengis.net/rdf#",
            "osmkey": "https://www.openstreetmap.org/wiki/Key:",
            "osm2rdfmember": "https://osm2rdf.cs.uni-freiburg.de/rdf/member#",
            "osmrel": "https://www.openstreetmap.org/relation/",
            "osm2rdf": "https://osm2rdf.cs.uni-freiburg.de/rdf/",
            "osm2rdf_geom": "https://osm2rdf.cs.uni-freiburg.de/rdf/geom#",
            "meta": "https://www.openstreetmap.org/meta/",
            "xsd": "http://www.w3.org/2001/XMLSchema#",
        }

    def testQueryGen(self):
        """
        test generating a query
        """
        query_name = "RelationExplore"
        param_dict = {"relid": "10492086"}
        endpoint = "osm-qlever"

        lod = self.query_handler.query_by_name(
            query_name=query_name, param_dict=param_dict, endpoint=endpoint
        )
        if self.debug:
            print(f"Query: {query_name}:")
            print(json.dumps(lod, indent=2))
        query_gen = QueryGen(self.prefixes)
        relid = param_dict["relid"]
        value = f"osmrel:{relid}"
        sparql_query = query_gen.gen(
            lod,
            main_var="rel",
            main_value=value,
            first_x=9,
            max_cardinality=1,
            comment_out=True,
        )

        if self.debug:
            print("Generated SPARQL Query:")
            print(sparql_query)

    def testQueryGenPrefix(self):
        """
        test queries to explore correct OSM Planet SPAQRL queries
        """
        query_gen = QueryGen(self.prefixes)
        for prefix, prop in [
            ("osmkey", "ref"),
            ("meta", "uid"),
            ("rdf", "type"),
        ]:
            with self.subTest(prefix=prefix, prop=prop):
                prefix_uri = self.prefixes[prefix]
                long_prop = f"{prefix_uri}{prop}"
                short_prop = query_gen.get_prefixed_property(long_prop)
                expected = f"{prefix}:{prop}"
                self.assertEqual(short_prop, expected)

    def testQueryGenSanitize(self):
        """
        Test QueryGen functions for correct prefix handling and variable sanitization.
        """
        query_gen = QueryGen(self.prefixes)
        expected_results = {
            "osmkey:ref": "ref",
            "meta:uid": "uid",
            "rdf:type": "type",
        }

        for prop, expected in expected_results.items():
            with self.subTest(prop=prop):
                sanitized = query_gen.sanitize_variable_name(prop)
                self.assertEqual(sanitized, expected)

    def test_gen_query(self):
        """
        Test SPARQL query generation.
        """
        for (
            title,
            main_var,
            prefix,
            node_id,
            endpoint_name,
            first_x,
            comment_out,
            expected_keys,
        ) in [
            (
                "Vía Verde (Burgos - Túnel de La Engaña)",
                "bike_route",
                "osmrel",
                "2172017",
                "osm-qlever",
                1000,
                False,
                24,
            ),
            (
                "MD 18061",
                "train_route",
                "osmrel",
                "10492086",
                "osm-qlever",
                1000,
                False,
                32,
            ),
            (
                "Christian Ronaldo",
                "person",
                "wd",
                "Q11571",
                "wikidata-qlever",
                5,
                False,
                1,
            ),
        ]:
            with self.subTest(node_id=node_id):
                explorer = Explorer(endpoint_name)
                prefixes = explorer.endpoint_prefixes.get(endpoint_name)
                query_gen = QueryGen(prefixes=prefixes, debug=self.debug)
                start_node = explorer.get_node(node_id, prefix)
                qlod = explorer.explore_node(
                    node=start_node, triple_pos=TriplePos.SUBJECT, summary=True
                )
                generated_query = query_gen.gen(
                    lod=qlod,
                    main_var=main_var,
                    main_value=start_node.qualified_name,
                    max_cardinality=1,
                    first_x=first_x,
                    comment_out=comment_out,
                )
                if self.debug:
                    print(generated_query)
                lod = explorer.query(
                    sparql_query=generated_query, param_dict={}, endpoint=endpoint_name
                )
                # we expect a single but long record
                self.assertEqual(len(lod), 1)
                record = lod[0]
                if self.debug:
                    print(title)
                    print(json.dumps(record, indent=2, default=str))
                self.assertEqual(len(record.keys()), expected_keys)



"""
Created on 2025-02-06

@author: wf
"""

import logging
import re
from pathlib import Path

from lodstorage.query import EndpointManager, Query, QueryManager
from lodstorage.sparql import SPARQL, Params


class NPQ_Handler:
    """
    Handling of named parameterized queries
    """

    def __init__(self, yaml_file: str, with_default: bool = False, debug: bool = False):
        """
        Constructor

        Args:
            yaml_file (str): The YAML file containing the queries.
            with_default (bool): Whether to include default endpoints.
            debug(bool): if True switch on debug mode
        """

        self.debug = debug
        self.endpoint_path = Path(__file__).parent / "resources" / "endpoints.yaml"
        self.query_path = Path(__file__).parent / "resources" / "queries"
        self.query_yaml = self.query_path / yaml_file

        if not self.query_yaml.is_file():
            raise FileNotFoundError(f"Queries file not found: {self.query_yaml}")

        self.query_manager = QueryManager(
            lang="sparql", queriesPath=self.query_yaml.as_posix()
        )

        self.endpoints = EndpointManager.getEndpoints(
            self.endpoint_path.as_posix(), with_default=with_default
        )

        # Preload prefixes for each endpoint
        self.endpoint_prefixes = {}
        for endpoint_name, endpoint in self.endpoints.items():
            if hasattr(endpoint, "prefixes"):
                prefix_dict, _body = self.parse_prefixes(endpoint.prefixes)
                self.endpoint_prefixes[endpoint_name] = prefix_dict
            else:
                self.endpoint_prefixes[endpoint_name] = dict()

    def parse_prefixes(self, prefix_str: str) -> dict:
        """
        Parse prefixes from string with newlines and prefixes per
        line into a dictionary

        Args:
            prefix_str (str): The string containing prefix definitions.

        Returns:
            dict: A dictionary mapping prefix names to their URIs.
        """
        prefix_pattern = re.compile(
            r"^prefix\s+(?P<name>\w+):\s+<(?P<uri>[^>]+)>", re.IGNORECASE
        )
        prefix_dict = {}
        body = ""
        for line in prefix_str.splitlines():
            if match := prefix_pattern.match(line.strip()):
                name = match.group("name")
                uri = match.group("uri")
                prefix_dict[name] = uri
            else:
                body += line + "\n"
        return prefix_dict, body

    def to_set(self, prefix_dict) -> set:
        prefix_set = set()
        for name in prefix_dict.keys():
            prefix_set.add(name)
        return prefix_set

    def merge_prefixes_by_endpoint_name(
        self, query_str: str, endpoint_name: str
    ) -> str:
        """
        Merge query prefixes with endpoint prefixes avoiding duplicates.

        Args:
            query_str (str): SPARQL query string potentially containing prefixes.
            endpoint_name (str): Name of the endpoint to use for prefix lookup.

        Returns:
            str: Query with merged unique prefixes.
        """
        endpoint_prefix_dict = self.endpoint_prefixes.get(endpoint_name, {})
        merged_query = self.merge_prefixes(query_str, endpoint_prefix_dict)
        return merged_query

    def merge_prefixes(self, query_str: str, endpoint_prefix_dict: dict) -> str:
        """
        Merge query prefixes with endpoint prefixes avoiding duplicates.

        Args:
            query_str (str): SPARQL query string potentially containing prefixes.
            endpoint_prefix_dict (dict): Dictionary mapping prefix names to URIs.

        Returns:
            str: Query with merged unique prefixes.
        """
        prefix_dict, body_section = self.parse_prefixes(query_str)
        query_prefix_set = self.to_set(prefix_dict)

        endpoint_prefix_set = self.to_set(endpoint_prefix_dict)

        missing_prefixes = endpoint_prefix_set - query_prefix_set

        if not missing_prefixes:
            return query_str

        merged_prefix_dict = dict(endpoint_prefix_dict)
        merged_prefix_dict.update(prefix_dict)

        merged_prefix_lines = []
        for prefix in sorted(merged_prefix_dict.keys()):
            uri = merged_prefix_dict[prefix]
            merged_prefix_lines.append(f"PREFIX {prefix}: <{uri}>")

        merged_query = "\n".join(merged_prefix_lines) + f"\n\n{body_section}"
        return merged_query

    def query_by_name(
        self,
        query_name: str,
        param_dict: dict = {},
        endpoint: str = "wikidata-qlever",
        auto_prefix: bool = True,
    ):
        """
        Get the result of the given query.

        Args:
            query_name (str): Name of the query to execute.
            param_dict (dict): Dictionary of parameters to substitute.
            endpoint (str): Name of the endpoint to use.
            auto_prefix (bool): Whether to automatically add endpoint prefixes.

        Returns:
            list: List of dictionaries with query results.
        """
        query: Query = self.query_manager.queriesByName.get(query_name)
        if not query:
            raise ValueError(f"{query_name} is not defined!")

        # Get the query string and handle prefixes
        sparql_query = query.query
        lod = self.query(
            sparql_query=sparql_query,
            param_dict=param_dict,
            endpoint=endpoint,
            auto_prefix=auto_prefix,
        )
        return lod

    def query(
        self,
        sparql_query: str,
        param_dict: dict = {},
        endpoint: str = "wikidata-qlever",
        auto_prefix: bool = True,
    ):
        """
        Get the result of the given query.

        Args:
            sparql_query (str): the query to execute.
            param_dict (dict): Dictionary of parameters to substitute.
            endpoint (str): Name of the endpoint to use.
            auto_prefix (bool): Whether to automatically add endpoint prefixes.

        Returns:
            list: List of dictionaries with query results.
        """

        sparql_endpoint = self.endpoints[endpoint]
        endpoint_instance = SPARQL(sparql_endpoint.endpoint)

        if auto_prefix:
            logging.debug(f"Auto prefixing for endpoint: {endpoint}")
            sparql_query = self.merge_prefixes_by_endpoint_name(sparql_query, endpoint)
        msg = f"SPARQL query:\n{sparql_query}"
        logging.debug(msg)
        if self.debug:
            print(msg)
            params = Params(sparql_query)
            final_query = params.apply_parameters_with_check(param_dict)
            print("parameterized query:")
            print(final_query)

        # Execute query
        lod = endpoint_instance.queryAsListOfDicts(sparql_query, param_dict=param_dict)
        return lod



"""
Created on 2025-01-19

@author: wf

Tour definition as Legs connecting Locs
Styling of Legs

"""

from dataclasses import field
from typing import Dict, Optional, Tuple

from basemkit.yamlable import lod_storable


@lod_storable
class Loc:
    """
    A location in a trip or tour
    """

    id: str
    coordinates: Tuple[float, float]  # latlon
    name: str
    loc_type: Optional[str] = None
    url: Optional[str] = None
    notes: Optional[str] = None


@lod_storable
class Leg:
    """
    A segment of a trip between two locations
    """

    leg_type: str  # e.g., "bike", "train", "car"
    start: Loc
    end: Loc
    url: Optional[str] = None


@lod_storable
class Tour:
    """
    A sequence of legs connecting waypoints that form a complete journey
    """

    name: str
    legs: list[Leg] = field(default_factory=list)
    description: Optional[str] = None
    url: Optional[str] = None

    def dump(self, limit: int = 10, leg_styles: "LegStyles" = None):
        """
        Print a detailed dump of the tour for debugging

        Args:
           limit: Maximum number of legs to show
        """
        if leg_styles is None:
            leg_styles = LegStyles.default()
        print(f"Tour: {self.name}")
        for i, leg in enumerate(self.legs):
            if i >= limit:
                remaining = len(self.legs) - limit
                print(f"... {remaining} more legs")
                break
            leg_style = leg_styles.get_style(leg.leg_type)
            if leg_style:
                utf8_icon = leg_style.utf8_icon
            else:
                utf8_icon = "?"
            coord_start = (
                f"({leg.start.coordinates[0]:.5f}, {leg.start.coordinates[1]:.5f})"
            )
            coord_end = f"({leg.end.coordinates[0]:.5f}, {leg.end.coordinates[1]:.5f})"
            print(f" {utf8_icon} {coord_start} ➜ {coord_end}")


@lod_storable
class LegStyle:
    """
    Style configuration for a transport leg
    """

    leg_type: str  # e.g. "bike", "train", "car", "ferry", "bus", "plane"
    point_type: str
    color: str
    utf8_icon: str
    weight: int
    dashArray: Optional[str]
    opacity: float


@lod_storable
class LegStyles:
    """
    Collection of predefined styles for different leg types
    """

    styles: Dict[str, LegStyle] = field(default_factory=dict)

    def get_style(self, leg_type: str) -> LegStyle:
        """
        Get style for given leg type
        """
        leg_style = self.styles.get(leg_type)
        return leg_style

    @classmethod
    def default(cls) -> "LegStyles":
        """
        Get default leg styles
        """
        default_styles = {
            "bike": LegStyle(
                leg_type="bike",
                point_type="knooppunt",
                color="#FF0000",  # bike - red (avoid green due to map background)
                utf8_icon="🚲",
                weight=3,
                dashArray=None,
                opacity=1.0,
            ),
            "train": LegStyle(
                leg_type="train",
                point_type="train_station",
                color="#555555",  # train - dark gray (improves contrast over black)
                utf8_icon="🚂",
                weight=3,
                dashArray="10,10",
                opacity=1.0,
            ),
            "car": LegStyle(
                leg_type="car",
                point_type="parking",
                color="#404040",  # car - medium gray
                utf8_icon="🚗",
                weight=3,
                dashArray=None,
                opacity=1.0,
            ),
            "ferry": LegStyle(
                leg_type="ferry",
                point_type="ferry_terminal",
                color="#1E90FF",  # ferry - dodger blue for visibility
                utf8_icon="⛴️",
                weight=3,
                dashArray="15,10",
                opacity=0.8,
            ),
            "foot": LegStyle(
                leg_type="foot",
                point_type="waypoint",
                color="#FFD700",  #  yellow
                utf8_icon="👣",
                weight=2,
                dashArray="5,5",
                opacity=0.9,
            ),
            "bus": LegStyle(
                leg_type="bus",
                point_type="bus_stop",
                color="#FF4500",  # bus - orange-red for distinctiveness
                utf8_icon="🚌",
                weight=3,
                dashArray=None,
                opacity=1.0,
            ),
            "plane": LegStyle(
                leg_type="plane",
                point_type="airport",
                color="#4B0082",  # plane - indigo for uniqueness
                utf8_icon="✈️",
                weight=3,
                dashArray="20,10,5,10",
                opacity=0.7,
            ),
        }
        return cls(styles=default_styles)



"""
Created on 2024-01-03

@author: wf
"""

import asyncio
from typing import Callable

from ez_wikidata.wdsearch import WikidataSearch
from ngwidgets.lod_grid import ListOfDictsGrid
from ngwidgets.webserver import WebSolution
from ngwidgets.widgets import Lang, Link
from nicegui import ui


class WikidataItemSearch:
    """
    wikidata item search
    """

    def __init__(
        self, solution: WebSolution, record_filter: Callable = None, lang: str = "en"
    ):
        """
        Initialize the WikidataItemSearch with the given solution.

        Args:
            solution (WebSolution): The solution to attach the search UI.
            record_filter(Callable): callback for displayed found records
        """
        self.solution = solution
        self.lang = lang
        # Get available languages
        self.languages = Lang.get_language_dict()
        self.record_filter = record_filter
        self.limit = 9
        self.wd_search = WikidataSearch(lang)
        self.search_debounce_task = None
        self.keyStrokeTime = 0.65  # minimum time in seconds to wait between keystrokes before starting searching
        self.search_result_row = None
        self.setup()

    def setup(self):
        """
        setup the user interface
        """
        with ui.card().style("width: 25%"):
            with ui.grid(rows=1, columns=4):
                # Create a label to display the chosen language with the default language
                self.lang_label = ui.label(self.lang)
                # Create a dropdown for language selection with the default language selected
                # Bind the label text to the selection's value, so it updates automatically
                ui.select(self.languages, value=self.lang).bind_value(
                    self.lang_label, "text"
                )
                ui.label("limit:")
                self.limit_slider = (
                    ui.slider(min=2, max=50, value=self.limit)
                    .props("label-always")
                    .bind_value(self, "limit")
                )
            with ui.row():
                self.search_input = ui.input(
                    label="search", on_change=self.on_search_change
                ).props("size=80")
        with ui.row() as self.search_result_row:
            self.search_result_grid = ListOfDictsGrid()

    async def on_search_change(self, _args):
        """
        react on changes in the search input
        """
        # Cancel the existing search task if it's still waiting
        if self.search_debounce_task:
            self.search_debounce_task.cancel()

        # Create a new task for the new search
        self.search_debounce_task = asyncio.create_task(self.debounced_search())

    async def debounced_search(self):
        """
        Waits for a period of inactivity and then performs the search.
        """
        try:
            # Wait for the debounce period (keyStrokeTime)
            await asyncio.sleep(self.keyStrokeTime)
            search_for = self.search_input.value
            if self.search_result_row:
                with self.search_result_row:
                    lang = self.lang_label.text
                    ui.notify(f"searching wikidata for {search_for} ({lang})...")
                    self.wd_search.language = lang
                    wd_search_result = self.wd_search.searchOptions(
                        search_for, limit=self.limit
                    )
                    view_lod = self.get_selection_view_lod(wd_search_result)
                    self.search_result_grid.load_lod(view_lod)
                    # self.search_result_grid.set_checkbox_selection("#")
                    self.search_result_grid.update()
        except asyncio.CancelledError:
            # The search was cancelled because of new input, so just quietly exit
            pass
        except BaseException as ex:
            self.solution.handle_exception(ex)

    def get_selection_view_lod(self, wd_search_result: list) -> dict:
        """
        Convert the Wikidata search result list of dict to a selection.

        Args:
            wd_search_result (List[Dict[str, Any]]): The search results from Wikidata.

        Returns:
            List[Dict[str, Any]]: The list of dictionaries formatted for view.
        """
        view_lod = []
        for qid, itemLabel, desc in wd_search_result:
            url = f"https://www.wikidata.org/wiki/{qid}"
            link = Link.create(url, qid)
            row = {
                "#": len(view_lod) + 1,
                "qid": link,
                "label": itemLabel,
                "desc": desc,
            }
            if self.record_filter:
                self.record_filter(qid, row)
            view_lod.append(row)
        return view_lod



"""
Created on 2025-02-01

@author: wf
"""

import os
import re

from ez_wikidata.wdproperty import WikidataPropertyManager
from ngwidgets.input_webserver import InputWebserver, InputWebSolution
from ngwidgets.webserver import WebserverConfig
from ngwidgets.widgets import Lang, Link
from nicegui import Client, app, ui

from velorail.explore import Explorer, TriplePos
from velorail.explore_view import ExplorerView
from velorail.gpxviewer import GPXViewer
from velorail.locfind import LocFinder
from velorail.sso_users_solution import SsoSolution
from velorail.version import Version
from velorail.wditem_search import WikidataItemSearch


class VeloRailSolution(InputWebSolution):
    """
    the VeloRail solution
    """

    def __init__(self, webserver: "VeloRailWebServer", client: Client):
        """
        Initialize the solution

        Calls the constructor of the base solution
        Args:
            webserver (VeloRailWebServer): The webserver instance associated with this context.
            client (Client): The client instance this context is associated with.
        """
        super().__init__(webserver, client)  # Call to the superclass constructor
        self.args = self.webserver.args
        self.lang = self.args.lang
        self.wpm = self.webserver.wpm
        self.viewer = GPXViewer(args=self.args)

    def configure_menu(self):
        """
        configure my menu
        """
        InputWebSolution.configure_menu(self)
        self.sso_solution = SsoSolution(webserver=self.webserver)
        self.sso_solution.configure_menu()

    def clean_smw_artifacts(self, input_str: str) -> str:
        """
        Remove SMW artifacts ([[SMW::on]] and [[SMW::off]]) from the input string.

        Args:
            input_str (str): Input string containing SMW artifacts.

        Returns:
            str: Cleaned string without SMW markers.
        """
        # Regex to match and remove SMW markers
        return re.sub(r"\[\[SMW::(on|off)\]\]", "", input_str)

    async def show_wikidata_item(self, qid: str = None):
        """
        show the given wikidata item on the map
        Args:
            qid(str): the Wikidata id of the item to analyze
        """

        def show():
            viewer = self.viewer
            # Create LocFinder and get coordinates
            locfinder = LocFinder()
            center = None
            wd_item = locfinder.get_wikidata_geo(qid)
            if wd_item:
                wd_link = wd_item.as_wd_link()
                wd_maps = wd_item.get_map_links(zoom=self.viewer.zoom)
                # create markup with links
                markup = f"{wd_link}&nbsp;{wd_maps}"
                ui.html(markup)
                center = [wd_item.lat, wd_item.lon]
            viewer.show(center=center)

        await self.setup_content_div(show)

    async def show_lines(
        self,
        lines: str = None,
        auth_token: str = None,
        zoom: int = GPXViewer.default_zoom,
    ):
        """
        Endpoint to display routes based on 'lines' parameter.
        """
        if not self.viewer:
            ui.label("Error: Viewer not initialized")
            return

        if self.viewer.args.token and auth_token != self.viewer.args.token:
            ui.label("Error: Invalid authentication token")
            return

        if not lines:
            ui.label("Error: No 'lines' parameter provided")
            return

        # Clean the lines parameter to remove SMW artifacts
        cleaned_lines = self.clean_smw_artifacts(lines)

        # Delegate logic to GPXViewer
        try:
            self.viewer.parse_lines_and_show(cleaned_lines, zoom=zoom)
        except ValueError as e:
            ui.label(f"Error processing lines: {e}")

    async def show_gpx(
        self,
        gpx: str = None,
        auth_token: str = None,
        zoom: int = GPXViewer.default_zoom,
    ):
        """
        GPX viewer page with optional gpx_url and auth_token.
        """
        viewer = self.viewer
        if not viewer:
            ui.label("Error: Viewer not initialized")
            return

        if viewer.args.token and auth_token != viewer.args.token:
            ui.label("Error: Invalid authentication token")
            return

        gpx_to_use = gpx if gpx else viewer.args.gpx
        if gpx_to_use:
            viewer.load_gpx(gpx_to_use)
            viewer.show(zoom=zoom)
        else:
            ui.label(
                "Please provide a GPX file via 'gpx' query parameter or the command line."
            )

    async def show_explorer(
        self,
        node_id: str = None,
        prefix: str = "osm",
        endpoint_name: str = "osm-qlever",
        summary: bool = False,
    ):
        """
        show the SPARQL explorer for the given node

        Args:
            node_id(str): id of the node to explore
            prefix(str): prefix to use e.g. wd:
            endpoint_name(str): name of the endpoint to use
            summary(bool): if True show summary
        """

        def show():
            explorer_view = ExplorerView(
                self, prefix=prefix, endpoint_name=endpoint_name, summary=summary
            )
            explorer_view.setup_ui()
            explorer_view.show(node_id)

        await self.setup_content_div(show)

    def prepare_ui(self):
        """
        overrideable configuration
        """
        self.endpoint_name = self.args.endpointName

    def configure_settings(self):
        """
        configure settings
        """
        lang_dict = Lang.get_language_dict()
        self.add_select("language:", lang_dict).bind_value(self, "lang")

    async def home(self):
        """
        provide the main content page
        """

        def record_filter(qid: str, record: dict):
            """
            filter the given search record
            """
            if "label" and "desc" in record:
                desc = record["desc"]
                label = record["label"]
                text = f"""{label}({qid})☞{desc}"""
                map_link = Link.create(f"/wd/{qid}", text)
                # getting the link to be at second position
                # is a bit tricky
                temp_items = list(record.items())
                # Add the new item in the second position
                temp_items.insert(1, ("map", map_link))

                explore_link = Link.create(
                    f"/explore/{qid}?prefix=wd&endpoint_name=wikidata&summary=True",
                    text,
                )
                temp_items.insert(2, ("explore", explore_link))

                # Clear the original dictionary and update it with the new order of items
                record.clear()
                record.update(temp_items)

        def show():
            self.wd_item_search = WikidataItemSearch(
                self, record_filter=record_filter, lang=self.lang
            )

        await self.setup_content_div(show)


class VeloRailWebServer(InputWebserver):
    """WebServer class that manages the server for velorail"""

    @classmethod
    def get_config(cls) -> WebserverConfig:
        copy_right = "(c)2025 velorail team"
        config = WebserverConfig(
            copy_right=copy_right,
            version=Version(),
            default_port=9876,
            short_name="velorail",
        )
        server_config = WebserverConfig.get(config)
        server_config.solution_class = VeloRailSolution
        return server_config

    def __init__(self):
        """Constructs all the necessary attributes for the WebServer object."""
        InputWebserver.__init__(self, config=VeloRailWebServer.get_config())

        # Initialize property manager instance
        self.wpm = WikidataPropertyManager.get_instance()

        @ui.page("/explore/{node_id}")
        async def explorer_page(
            client: Client,
            node_id: str,
            prefix: str = "osmrel",
            endpoint_name: str = "osm-qlever",
            summary: bool = False,
        ):
            """
            explore the given node id
            """
            await self.page(
                client,
                VeloRailSolution.show_explorer,
                node_id=node_id,
                prefix=prefix,
                endpoint_name=endpoint_name,
                summary=summary,
            )

        @ui.page("/wd/{qid}")
        async def wikidata_item_page(client: Client, qid: str):
            """
            show the given wikidata item on the map
            """
            await self.page(client, VeloRailSolution.show_wikidata_item, qid)

        @ui.page("/lines")
        async def lines_page(
            client: Client,
            lines: str = None,
            auth_token: str = None,
            zoom: int = GPXViewer.default_zoom,
        ):
            """
            Endpoint to display routes based on 'lines' parameter.
            """
            await self.page(
                client, VeloRailSolution.show_lines, lines, auth_token, zoom
            )

        @ui.page("/gpx")
        async def gpx_page(
            client: Client,
            gpx: str = None,
            auth_token: str = None,
            zoom: int = GPXViewer.default_zoom,
        ):
            """
            GPX viewer page with optional gpx_url and auth_token.
            """
            await self.page(client, VeloRailSolution.show_gpx, gpx, auth_token, zoom)

        @app.get("/api/explore/{node_id}")
        async def explore_api(
            node_id: str,
            prefix: str = "osmrel",
            endpoint_name: str = "osm-qlever",
            summary: bool = False,
        ):
            """
            SPARQL explorer REST API endpoint

            Args:
                node_id: id of the node to explore
                prefix: prefix to use e.g. wd:
                endpoint_name: name of the endpoint to use
                summary: if True show summary

            Returns:
                dict: JSON response with exploration results
            """
            explorer = Explorer(endpoint_name)

            start_node = explorer.get_node(prefix=prefix, node_id=node_id)
            try:
                lod = explorer.explore_node(
                    start_node, triple_pos=TriplePos.SUBJECT, summary=summary
                )
                return {"status": "ok", "records": lod}
            except Exception as ex:
                return {"status": "error", "message": str(ex)}

    def configure_run(self):
        root_path = (
            self.args.root_path
            if self.args.root_path
            else VeloRailWebServer.examples_path()
        )
        self.root_path = os.path.abspath(root_path)
        self.allowed_urls = [
            "https://raw.githubusercontent.com/WolfgangFahl/velorail/main/velorail_examples/",
            self.examples_path(),
            self.root_path,
        ]

    @classmethod
    def examples_path(cls) -> str:
        # the root directory (default: examples)
        path = os.path.join(os.path.dirname(__file__), "../velorail_examples")
        path = os.path.abspath(path)
        return path



"""
Created on 2025-02-01

@author: th
"""

from typing import Optional

import numpy as np
import pandas as pd
from basemkit.yamlable import lod_storable
from ngwidgets.widgets import Link

from velorail.npq import NPQ_Handler
from velorail.tour import LegStyles


@lod_storable
class WikidataGeoItem:
    """
    Dataclass for storing Wikidata geographical location data with labels
    """

    qid: str
    lat: float
    lon: float
    label: Optional[str] = None
    description: Optional[str] = None

    def as_wd_link(self) -> Link:
        text = f"""{self.label}({self.qid})☞{self.description}"""
        wd_link = Link.create(f"https://www.wikidata.org/wiki/{self.qid}", text)
        return wd_link

    def get_map_links(
        self, leg_styles: Optional[LegStyles] = None, zoom: int = 14
    ) -> str:
        """
        Get HTML markup with icons grouped by map type
        """
        if leg_styles is None:
            leg_styles = LegStyles.default()

        map_links = {
            "openstreetmap.org": ["car", "bus", "plane"],
            "opencyclemap.org": ["bike"],
            "openrailwaymap.org": ["train"],
            "map.openseamap.org": ["ferry"],
            "hiking.waymarkedtrails.org": ["foot"],
        }

        markup = ""
        delim = ""
        for map_url, leg_types in map_links.items():
            icons = ""
            for leg_type in leg_types:
                leg_style = leg_styles.get_style(leg_type)
                icons += leg_style.utf8_icon
            tooltip = f"{','.join(leg_types)} map"
            if "car" in leg_types:
                url = f"https://{map_url}/#map={zoom}/{self.lat}/{self.lon}"
            elif "foot" in leg_types:
                url = f"https://{map_url}/#?map={zoom}/{self.lat}/{self.lon}"
            else:
                url = f"https://{map_url}/?zoom={zoom}&lat={self.lat}&lon={self.lon}"
            link = Link.create(url, text=icons, tooltip=tooltip, target="_blank")
            markup += link + delim
            delim = "\n"
        return markup

    @property
    def osm_url(self, map_type: str = "street", zoom: int = 15) -> str:
        """
        Get OpenStreetMap URL for this location

        Args:
            zoom: Zoom level (default=15)

        Returns:
            OpenStreetMap URL for the location
        """
        osm_url = f"https://www.open{map_type}map.org/?mlat={self.lat}&mlon={self.lon}&zoom={zoom}"
        return osm_url

    @classmethod
    def from_record(cls, record: dict) -> "WikidataGeoItem":
        """
        Create WikidataGeoItem from a dictionary record

        Args:
            record: Dictionary containing lat, lon, label and description

        Returns:
            WikidataGeoRecord instance
        """
        return cls(
            qid=record["qid"],
            lat=float(record["lat"]),
            lon=float(record["lon"]),
            label=record["label"],
            description=record["description"],
        )


class LocFinder(NPQ_Handler):
    """
    Set of methods to lookup different location types
    """

    def __init__(self):
        super().__init__("locations.yaml")

    def get_bike_nodes_by_bounds(
        self, south: float, west: float, north: float, east: float
    ):
        """
        Get all bike nodes within the given bounding box

        Args:
            south: Southern latitude boundary
            west: Western longitude boundary
            north: Northern latitude boundary
            east: Eastern longitude boundary

        Returns:
            list of dicts containing bike route information
        """
        param_dict = {"south": south, "west": west, "north": north, "east": east}
        lod = self.query_by_name(
            query_name="BikeNodes4Bounds",
            param_dict=param_dict,
            endpoint="osm-sophox",  # Using Sophox endpoint for OSM data
        )
        return lod

    def get_wikidata_geo(self, qid: str) -> WikidataGeoItem:
        """
        Get geographical coordinates and metadata for a Wikidata item

        Args:
            qid: Wikidata QID of the item

        Returns:
            WikidataGeoItem with location data and metadata
        """
        lod = self.query_by_name(query_name="WikidataGeo", param_dict={"qid": qid})
        if len(lod) >= 1:
            record = lod[0]
            record["qid"] = qid  # Add qid to record for WikidataGeoItem creation
            return WikidataGeoItem.from_record(record)
        return None

    def get_all_train_stations(self):
        lod = self.query_by_name(query_name="AllTrainStations")
        return lod

    def get_train_stations_by_coordinates(
        self, latitude: float, longitude: float, radius: float
    ):
        """
        Get all train stations within the given radius around the given latitude and longitude
        """
        lod = self.get_all_train_stations()
        df = pd.DataFrame.from_records(lod)
        # Haversine formula components
        lat1, lon1 = np.radians(latitude), np.radians(longitude)
        lat2, lon2 = np.radians(df["lat"]), np.radians(df["long"])

        # Differences in coordinates
        dlat = lat2 - lat1
        dlon = lon2 - lon1

        # Haversine formula
        a = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
        c = 2 * np.arcsin(np.sqrt(a))
        r = 6371  # Radius of Earth in kilometers

        # Calculate distances
        distances = c * r

        # Add distances to dataframe
        df_with_distances = df.copy()
        df_with_distances["distance_km"] = distances

        # Filter points within radius
        points_within_radius = df_with_distances[
            df_with_distances["distance_km"] <= radius
        ].copy()

        # Sort by distance
        points_within_radius = points_within_radius.sort_values("distance_km")

        return points_within_radius



"""
Created on 2025-02-01

@author: th
"""

import json

from ngwidgets.basetest import Basetest

from velorail.locfind import LocFinder


class TestLocFinder(Basetest):
    """
    test locfinder
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_wikidata_loc(self):
        """
        test finding location of a wikidata item
        """
        locfinder = LocFinder()
        # Test with Gare de Biarritz (Q1959795)
        qid = "Q1959795"
        expected = {
            "lat": 43.4592,
            "lon": -1.5459,
            "label": "Gare de Biarritz",
            "description": "railway station in Biarritz, France",
        }

        lod = locfinder.query_by_name(query_name="WikidataGeo", param_dict={"qid": qid})
        self.assertTrue(len(lod) >= 1)
        record = lod[0]

        # Check all fields are present with expected values
        for key, expected_value in expected.items():
            self.assertIn(key, record)
            if isinstance(expected_value, float):
                self.assertAlmostEqual(float(record[key]), expected_value, places=3)
            else:
                self.assertEqual(record[key], expected_value)

    def test_wikidata_item_maps(self):
        """
        test WikidataGeoItem map link generation
        """
        locfinder = LocFinder()
        # Test with San Sebastian station (Q14314)
        qid = "Q14314"
        # Get the WikidataGeoItem
        wd_item = locfinder.get_wikidata_geo(qid)
        self.assertIsNotNone(wd_item)

        # Get map links
        map_links = wd_item.get_map_links()
        if self.debug:
            print(map_links)

    def test_get_train_stations(self):
        """
        test get_train_stations
        """
        locfinder = LocFinder()
        lod_train_stations = locfinder.get_all_train_stations()
        print(len(lod_train_stations))
        self.assertGreaterEqual(len(lod_train_stations), 70000)

    def test_get_nearest_train_station(self):
        """
        test get_nearest_train_station
        """
        lat = 43.2661645
        long = -1.9749167
        distance = 10
        locfinder = LocFinder()
        results = locfinder.get_train_stations_by_coordinates(lat, long, distance)
        print(results)
        self.assertGreaterEqual(len(results), 30)

    def test_get_bike_nodes_heeg_stavoren(self):
        """
        test getting bike nodes in the Heeg/Stavoren area of Friesland
        """
        # Bounding box for Heeg/Stavoren area
        south = 52.8349  # Southern boundary
        west = 5.3184  # Western boundary
        north = 53.0125  # Northern boundary
        east = 5.8279  # Eastern boundary

        locfinder = LocFinder()
        bike_routes = locfinder.get_bike_nodes_by_bounds(south, west, north, east)
        if self.debug:
            print(json.dumps(bike_routes, indent=2))



"""
Created on 2025-01-17

@author: wf
"""

from ngwidgets.basetest import Basetest

from velorail.gpxviewer import GPXViewer


class TestGPXViewer(Basetest):
    """
    test GPXViewer
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_gpx_viewer(self):
        """
        Test the GPX viewer functionality for all sample GPX files.
        Tests loading GPX files, tour creation, and center calculation.
        """
        sample_keys = list(GPXViewer.samples.keys())
        expected_centers = {
            sample_keys[0]: (  # Mountain bike loop at Middlesex Fells reservation.
                42.44903,
                -71.1139805,
            ),
            sample_keys[1]: (43.302147, -1.8539505),  # Via Verde de Arditurri
            sample_keys[2]: (42.89977, -2.34528),  # Vía Verde del Fc Vasco Navarro
        }

        for sample_name, gpx_url in GPXViewer.samples.items():
            with self.subTest(sample_name=sample_name):
                # Create viewer instance and load GPX file
                viewer = GPXViewer.from_url(gpx_url)

                viewer.set_center()

                if self.debug:
                    print(f"{sample_name}:{viewer.gpx}")
                    viewer.tour.dump()

                # Test GPX loaded successfully
                self.assertIsNotNone(viewer.gpx)

                # Test that tour and legs were created
                self.assertIsNotNone(viewer.tour)
                self.assertGreater(
                    len(viewer.tour.legs), 0, f"No legs found in {sample_name}"
                )

                # Test map properties
                self.assertEqual(viewer.zoom, GPXViewer.default_zoom)

                # Test center calculation
                exp_lat, exp_lon = expected_centers[sample_name]
                act_lat, act_lon = viewer.center

                self.assertAlmostEqual(
                    act_lat,
                    exp_lat,
                    places=5,
                    msg=f"Latitude mismatch for {sample_name}",
                )
                self.assertAlmostEqual(
                    act_lon,
                    exp_lon,
                    places=5,
                    msg=f"Longitude mismatch for {sample_name}",
                )

                # Test bounding box calculation
                self.assertIsNotNone(
                    viewer.bounding_box,
                    f"Bounding box not calculated for {sample_name}",
                )

                # Test map creation with leg styles
                viewer.show()
                self.assertIsNotNone(viewer.map, f"Map not created for {sample_name}")

    def test_lines_parsing(self):
        """
        Test parsing of lines input for routes.
        """
        input_lines = """
        By bike→
        51.243931° N, 6.520022° E,
        51.269222° N, 6.625467° E:
        By train→
        51.269222° N, 6.625467° E,
        51.220189° N, 6.792675° E:
        By train→
        51.220189° N, 6.792675° E,
        49.600136° N, 6.134017° E:
        By train→
        49.600136° N, 6.134017° E,
        48.689858° N, 6.175906° E
        """

        # Expected parsed legs
        expected_legs = [
            ("bike", (51.243931, 6.520022), (51.269222, 6.625467)),
            ("train", (51.269222, 6.625467), (51.220189, 6.792675)),
            ("train", (51.220189, 6.792675), (49.600136, 6.134017)),
            ("train", (49.600136, 6.134017), (48.689858, 6.175906)),
        ]

        # Instantiate the GPXViewer and parse lines
        viewer = GPXViewer()
        tour = viewer.parse_lines(input_lines)
        viewer.set_center()
        if self.debug:
            viewer.tour.dump()

        # Verify tour name
        self.assertEqual(tour.name, "Parsed Tour")

        # Verify number of legs
        self.assertEqual(len(tour.legs), len(expected_legs))

        # Verify parsed legs
        for leg, (expected_type, expected_start, expected_end) in zip(
            tour.legs, expected_legs
        ):
            self.assertEqual(leg.leg_type, expected_type)
            self.assertEqual(leg.start.coordinates, expected_start)
            self.assertEqual(leg.end.coordinates, expected_end)



"""
Created on 2025-02-01

@author: wf
"""

import math
from dataclasses import dataclass
from typing import List, Tuple

from shapely.geometry import LineString, Point, Polygon
from shapely.wkt import loads


@dataclass
class WKT:
    """
    Utility class for handling WKT (Well Known Text) geometries
    """

    wkt: str  # Store WKT string as an instance attribute
    # reference point
    lat: float = None
    lon: float = None

    def to_latlon(self) -> Tuple[float, float]:
        """
        Converts a WKT string to (latitude, longitude) for a Point,
        or the centroid for LineString, Polygon, etc.

        Returns:
            tuple(float, float): (latitude, longitude)
        """
        geom = loads(self.wkt)
        center = geom if isinstance(geom, Point) else geom.centroid
        return center.y, center.x  # (lat, lon)

    def to_latlon_list(self) -> List[Tuple[float, float]]:
        """
        Extracts all (latitude, longitude) coordinate pairs from a WKT geometry.

        Returns:
            list of tuples [(lat, lon), ...] for LineString/Polygon
            or [(lat, lon)] for a Point
        """
        geom = loads(self.wkt)

        if isinstance(geom, Point):
            return [(geom.y, geom.x)]  # Ensure a list for consistency
        elif isinstance(geom, (LineString, Polygon)):
            return [
                (lat, lon) for lon, lat in geom.coords
            ]  # Extract coordinates correctly
        else:
            raise ValueError(f"Unsupported geometry type: {type(geom)}")

    def to_latlon_str(self, precision: int = 5) -> str:
        """
        Convert WKT to a lat,lon string with specified precision.

        Args:
            precision: Number of decimal places (default 5 for ~1m precision)

        Returns:
            str: Comma-separated lat,lon with specified precision
        """
        # if we have no reference point use the centroid
        if not self.lat:
            self.lat, self.lon = self.to_latlon()
        text = f"{self.lat:.{precision}f},{self.lon:.{precision}f}"
        return text

    @staticmethod
    def distance(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
        """
        Calculate the great circle distance between two points on earth using the haversine formula.

        Args:
            lat1: Latitude of first point in degrees
            lon1: Longitude of first point in degrees
            lat2: Latitude of second point in degrees
            lon2: Longitude of second point in degrees

        Returns:
            float: Distance between points in meters
        """
        R = 6371000  # Earth radius in meters
        dLat = math.radians(lat2 - lat1)
        dLon = math.radians(lon2 - lon1)
        lat1 = math.radians(lat1)
        lat2 = math.radians(lat2)

        a = (
            math.sin(dLat / 2.0) ** 2
            + math.cos(lat1) * math.cos(lat2) * math.sin(dLon / 2.0) ** 2
        )
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
        return R * c  # Distance in meters



"""
Created on 2025-02-12

@author: wf
"""
from ngwidgets.basetest import Basetest

from velorail.wkt import WKT


class TestWKT(Basetest):
    """
    Test WKT handling and conversions
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_wkt_to_latlon(self):
        """
        Test conversion of WKT to latlon or centroid
        """
        test_cases = [
            (
                "POINT(-3.319103 42.542723)",
                {"coords": (42.542723, -3.319103), "expected": ("42.542", "-3.319")},
            ),
            (
                "LINESTRING(-3.319103 42.542723, -3.319200 42.543000, -3.318900 42.543100)",
                {"coords": (42.542941, -3.319067), "expected": ("42.542", "-3.319")},
            ),
            (
                "POLYGON((-3.319103 42.542723, -3.319200 42.543000, -3.318900 42.543100, -3.319103 42.542723))",
                {"coords": (42.542941, -3.319067), "expected": ("42.542", "-3.319")},
            ),
        ]

        for wkt_str, test_data in test_cases:
            with self.subTest(wkt_str=wkt_str):
                wkt=WKT(wkt_str)
                lat, lon =wkt.to_latlon()
                e_lat, e_lon = test_data["coords"]
                e_lat_str, e_lon_str = test_data["expected"]
                self.assertAlmostEqual(lat, e_lat, places=4)
                self.assertAlmostEqual(lon, e_lon, places=4)
                latlon_str = wkt.to_latlon_str()
                if self.debug:
                    print(f"{latlon_str}")
                self.assertTrue(e_lat_str in latlon_str)
                self.assertTrue(e_lon_str in latlon_str)



"""
SPARQL Graph Explorer - A tool for interactive exploration of RDF graphs
Created on 2025-02-06
@author: wf
"""

import re
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from ez_wikidata.wdproperty import WikidataProperty, WikidataPropertyManager
from ngwidgets.widgets import Link

from velorail.npq import NPQ_Handler


class TriplePos(Enum):
    SUBJECT = "subject"
    PREDICATE = "predicate"
    OBJECT = "object"


@dataclass
class Node:
    """
    Represents a node in the RDF graph
    """

    prefix: str
    node_id: str
    uri: Optional[str] = None
    label: Optional[str] = None

    @property
    def qualified_name(self) -> str:
        """
        Returns the qualified name in the format {prefix}:{node_id}
        """
        return f"{self.prefix}:{self.node_id}"


class Explorer(NPQ_Handler):
    """
    A SPARQL explorer that allows traversing RDF graphs starting from any node
    """

    def __init__(self, endpoint_name: str):
        """
        Initialize the explorer with a SPARQL endpoint

        Args:
            endpoint_name: Name of the SPARQL endpoint to query as defined in endpoints.yaml
        """
        super().__init__("sparql-explore.yaml")
        self.endpoint_name = endpoint_name
        self.wpm = WikidataPropertyManager.get_instance()

    def get_prop(self, value: str) -> WikidataProperty:
        """
        Get the WikidataProperty for the given value string - either from a URI or direct property id

        Args:
            value (str): The value to check - either a URI or direct property id

        Returns:
            WikidataProperty: The property if found else None
        """
        prop = None
        if "wikidata" in self.endpoint_name and self.wpm:
            prop_pattern = r"P(\d+)"
            # Only try to extract pid if it's a property URI or matches P-number pattern
            if "www.wikidata.org/prop" in value or re.match(prop_pattern, value):
                pid = re.sub(f".*{prop_pattern}.*", r"P\1", value)
                if pid:
                    prop = self.wpm.get_property_by_id(pid)
        return prop

    def get_view_record(self, record: dict, index: int) -> dict:
        """
        get the view record for the given record

        Args:
            record (dict): The source record to convert
            index (int): The row index

        Returns:
            dict: The view record with formatted links and properties
        """
        view_record = {"#": index}  # Number first
        record_copy = record.copy()
        for key, value in record_copy.items():
            if isinstance(value, str):
                prop = self.get_prop(value)
                if prop:
                    view_record[key] = Link.create(
                        prop.url, f"{prop.plabel} ({prop.pid})"
                    )
                    continue
                if value.startswith("http"):
                    view_record[key] = Link.create(value, value)
            else:
                view_record[key] = value
        return view_record

    def get_node(self, node_id: str, prefix: str) -> Node:
        """
        Resolve a node URI using stored prefixes.

        Args:
            node_id (str): The node identifier.
            prefix (str): The prefix to resolve.

        Returns:
            Node: Constructed node with resolved URI.
        """
        endpoint_prefix_dict = self.endpoint_prefixes.get(self.endpoint_name, {})

        if prefix in endpoint_prefix_dict:
            base_uri = endpoint_prefix_dict[prefix]
            uri = f"{base_uri}{node_id}"
        else:
            raise ValueError(
                f"Prefix '{prefix}' not found in endpoint '{self.endpoint_name}'"
            )

        node = Node(
            uri=uri,
            prefix=prefix,
            node_id=node_id,
        )
        return node

    def explore_node(
        self, node: Node, triple_pos: TriplePos, summary: bool = False
    ) -> str:
        """
        Get the appropriate exploration query based on node type

        Args:
            node: The node to explore from
            triple_pos: The triple position
            summary: show a summary with counts

        Returns:
            Query result from the appropriate SPARQL query
        """
        query_map = {
            TriplePos.SUBJECT: (
                "ExploreFromSubject" if not summary else "ExploreFromSubjectSummary"
            ),
            TriplePos.PREDICATE: (
                "ExploreFromPredicate" if not summary else "ExploreFromPredicateSummary"
            ),
            TriplePos.OBJECT: (
                "ExploreFromObject" if not summary else "ExploreFromObjectSummary"
            ),
        }

        query_name = query_map[triple_pos]

        param_dict = {"start_node": node.qualified_name}

        lod = self.query_by_name(
            query_name=query_name, param_dict=param_dict, endpoint=self.endpoint_name
        )
        return lod



"""
Created on 15.02.2025

@author: wf
"""
import logging
import os
import shutil
import tempfile
import yaml
from ngwidgets.basetest import Basetest
from wikibot3rd.sso import SSO, User
from wikibot3rd.sso_users import Sso_Users
from velorail.sso_users_solution import SsoSolution
from unittest.mock import Mock, patch


class TestSsoAuth(Basetest):
    """
    test SSO Authentication
    """
    def setUp(self, debug=True, profile=True):
        """
        setup the test
        """
        Basetest.setUp(self, debug, profile)
        mock_ws = Mock()
        mock_ws.config = Mock()
        mock_ws.config.short_name = "test"
        self.mock_credentials(mock_ws)

    def tearDown(self):
        """
        cleanup temporary test directory
        """
        self.mock_credentials()
        Basetest.tearDown(self)

    def mock_credentials(self,webserver=None):
        """
        when called with webserver setup else tear down
        """
        def mocked_SSO_constructor(
               i_self,  # inner self for the patched SSO constructor
               host: str,
               database: str,
               sql_port: int = 3306,
               db_username: str = None,
               db_password: str = None,
               with_pool: bool = True,
               timeout: float = 3,
               debug: bool = False):
            # set properties
            i_self.host = host
            i_self.database = database
            i_self.sql_port = sql_port
            i_self.db_username = db_username
            i_self.db_password = db_password
            i_self.with_pool = with_pool
            i_self.timeout = timeout
            i_self.debug = debug
            # mock functions
            i_self.check_port = lambda: True
            i_self.get_user = lambda: self.mock_user
            i_self.check_credentials = lambda: True
            i_self.verify_password = lambda: True
            i_self.fetch_user_data_from_database = lambda: None
            msg=f"providing mocked SSO {i_self.__dict__} with check_port, get_user, ... returning default values"
            if self.debug:
                logging.warn(msg)
            else:
                logging.info(msg)
        if not webserver:
            if os.path.exists(self.test_dir):
                shutil.rmtree(self.test_dir)
        else:
            # Create temp test directory with credentials
            test_dir = tempfile.mkdtemp()
            solutions_path = os.path.join(test_dir, ".solutions", "test")
            os.makedirs(solutions_path)

            # Create minimal credentials file
            creds = {
                "server_url": "http://test-server",
                "username": "test-user",
                "password": "test-password",
                "secret": "test-secret",
                "wiki_id": "test-wiki"
            }
            self.credentials_path=os.path.join(solutions_path, "sso_credentials.yaml")
            with open(self.credentials_path, "w") as f:
                yaml.dump(creds, f)

            self.test_dir = test_dir
            # lets mock the SSO and user handling
            # from wikibot3rd to avoid actually creating
            # a database connection to the target wiki
            # mock a user (as would be returned from a
            # row in the mediawiki user table)
            mock_user = Mock(spec=User)
            mock_user.id = 1
            mock_user.name = "test_user"
            mock_user.is_admin = False
            # mock the Mediawiki Single Signon
            # which behaves as if we did a successful
            # remote MySQL DB access and port checking
            # (which we absolutely must avoid in the test environment!
            self.mock_sso = Mock(spec=SSO)
            with patch.object(SSO, "__init__", mocked_SSO_constructor):
                self.sso_auth = SsoSolution(webserver, credentials_path=self.credentials_path)
                self.sso_auth.users = Sso_Users(solution_name="test", debug=self.debug, credentials_path=self.credentials_path)

    def test_get_user_display_name(self):
        """
        test the user display name handling with different cases
        """
        test_cases = [
            # username, expected_display, is_admin, is_available, msg
            (None, "?", False, True, "sso available but no user"),
            ("test_user", "test_user", False, True, "regular user logged in"),
            ("admin_user", "admin_user🔑", True, True, "admin user logged in"),
        ]

        for username, expected_display, is_admin, is_available, msg in test_cases:
            with self.subTest(msg=msg):

                # Verify SSO availability
                self.assertEqual(self.sso_auth.users.is_available, is_available,
                                 f"Expected is_available={is_available}, but got {self.sso_auth.users.is_available}")

                # Setup login mock
                mock_login = Mock()
                self.sso_auth.login = mock_login
                mock_login.get_username.return_value = username

                if username and is_available:
                    mock_user = Mock()
                    mock_user.is_admin = is_admin
                    self.sso_auth.get_user = Mock(return_value=mock_user)

                user_display = self.sso_auth.get_user_display_name()
                if self.debug:
                    print(user_display)
                self.assertEqual(expected_display, user_display)



"""
Created on 2025-06-02

@author: wf
"""

import json
import logging

from ngwidgets.basetest import Basetest

from velorail.explore import Explorer, Node, TriplePos
from velorail.npq import NPQ_Handler


class TestExplorer(Basetest):
    """
    test explorer
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        if debug:
            logging.basicConfig(level=logging.DEBUG)
        self.handler = NPQ_Handler("sparql-explore.yaml")

    def test_explorer(self):
        """
        Test explorer with different endpoints and examples
        """
        # Define test cases for different endpoints
        endpoint_tests = {
            "osm": {
                "endpoints": ["osm-qlever", "osm-sophox"],
                "examples": {"cycle_route": {"prefix": "osmrel", "id": "10492086"}},
            },
            "wikidata": {
                "endpoints": ["wikidata", "wikidata-qlever"],
                "examples": {"Tim Berners-Lee": {"prefix": "wd", "id": "Q80"}},
            },
        }

        for _platform, config in endpoint_tests.items():
            for endpoint_name in config["endpoints"]:
                explorer = Explorer(endpoint_name)
                self.assertIsNotNone(explorer)

                # Test each example for this endpoint
                for example_name, example in config["examples"].items():
                    if self.debug:
                        print(f"\nTesting {example_name} on {endpoint_name}")

                    # Create start node
                    prefix = example["prefix"]
                    node_id = example["id"]
                    start_node = explorer.get_node(node_id=node_id, prefix=prefix)
                    # Get exploration results
                    for summary in (False, True):
                        lod = explorer.explore_node(
                            start_node, triple_pos=TriplePos.SUBJECT, summary=summary
                        )
                        if self.debug:
                            print(f"{len(lod)}")
                            debug_limit = 7 if summary else 9
                            debug_lod = lod[:debug_limit]
                            print(json.dumps(debug_lod, indent=2, default=str))

    def test_merge_prefixes(self):
        """
        Test merging of prefixes from different sources
        """
        # Test case 1: No duplicates
        query = """PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT ?s ?p ?o
WHERE { ?s ?p ?o }"""

        endpoint_prefixes = """PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>"""
        endpoint_prefix_dict, _body = self.handler.parse_prefixes(endpoint_prefixes)
        merged = self.handler.merge_prefixes(query, endpoint_prefix_dict)
        self.assertTrue("PREFIX rdfs:" in merged)
        self.assertTrue("PREFIX owl:" in merged)
        self.assertTrue("PREFIX xsd:" in merged)

        # Test case 2: With duplicates
        query2 = """PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT ?s ?p ?o
WHERE { ?s ?p ?o }"""

        merged2 = self.handler.merge_prefixes(query2, endpoint_prefix_dict)
        # Count occurrences of each prefix
        self.assertEqual(merged2.count("PREFIX owl:"), 1)
        self.assertEqual(merged2.count("PREFIX rdfs:"), 1)

        # Test case 3: Empty endpoint prefixes
        merged3 = self.handler.merge_prefixes(query, {})
        self.assertEqual(merged3, query)

        # Test case 4: Query without prefixes
        query4 = "SELECT ?s ?p ?o WHERE { ?s ?p ?o }"
        merged4 = self.handler.merge_prefixes(query4, endpoint_prefix_dict)
        self.assertTrue("PREFIX owl:" in merged4)
        self.assertTrue("PREFIX xsd:" in merged4)
        self.assertTrue("SELECT ?s" in merged4)



#!/usr/bin/env python
import json
import os
import dataclasses
from argparse import ArgumentParser, Namespace
from typing import Dict, List

from lodstorage.query_cmd import QueryCmd

from velorail.npq import NPQ_Handler
from velorail.tour import LegStyles
from velorail.wkt import WKT


class Osm2WikiConverter:
    """
    Converter for OSM relations and nodes to MediaWiki pages
    """

    def __init__(self, args: Namespace):
        """
        Initialize the converter

        Args:
             args: command line args
        """
        self.args = args
        self.tmpdir = args.tmp
        self.query_handler = NPQ_Handler("osmplanet.yaml", debug=args.debug)
        self.leg_style = LegStyles.default()
        self.test = False

    @classmethod
    def get_parser(cls):
        """Get the argument parser"""
        parser = ArgumentParser(
            description="Convert OpenStreetMap relations and nodes to MediaWiki pages"
        )
        # Add standard query command args
        QueryCmd.add_args(parser)
        # Add our specific args
        parser.add_argument(
            "--tmp", default="/tmp", help="Temporary directory (default: %(default)s)"
        )
        parser.add_argument(
            "-en",
            "--endpoint_name",
            default="osm-qlever",
            help="Endpoint name (default: %(default)s)",
        )
        parser.add_argument(
            "--zoom", type=int, default=8, help="Zoom factor (default: %(default)s)"
        )
        parser.add_argument(
            "--min_lat",
            type=float,
            default=42.0,
            help="Minimum latitude (default: %(default)s)",
        )
        parser.add_argument(
            "--max_lat",
            type=float,
            default=44.0,
            help="Maximum latitude (default: %(default)s)",
        )
        parser.add_argument(
            "--min_lon",
            type=float,
            default=-9.0,
            help="Minimum longitude (default: %(default)s)",
        )
        parser.add_argument(
            "--max_lon",
            type=float,
            default=4.0,
            help="Maximum longitude (default: %(default)s)",
        )
        parser.add_argument(
            "--min_node_distance",
            type=float,
            default=1000,
            help="Minimum distance between nodes in m (default: 1000)",
        )
        parser.add_argument(
            "--role",
            default="stop",
            help="Member role to filter (default: %(default)s)",
        )
        parser.add_argument(
            "--country",
            default="Spanien",
            help="Country name for Loc template (default: %(default)s)",
        )
        parser.add_argument(
            "--category",
            default="Spain2025",
            help="Wiki category (default: %(default)s)",
        )
        parser.add_argument(
            "--loc_type",
            default="osm_node",
            help="location type (default: %(default)s)",
        )
        parser.add_argument(
            "--transport",
            default="bike",
            help="transport (default: %(default)s)",
        )
        parser.add_argument(
            "--osm_items",
            nargs="*",
            default=[
                "relation/10492086",
                "relation/4220975",
                "relation/1713826",
                "node/11757382798",
            ],
            help="osm items to process [default: %(default)s]",
        )
        args = parser.parse_args()
        return args

    def query_osm_item(self, osm_item: str) -> Dict:
        """
        Query the given osm_item using SPARQL

        Args:
             osm_item: The osm_item to query

        Returns:
             Dict: The query results as list of dicts
        """
        if osm_item.startswith("relation"):
            queryName = "ItemNodesGeo"
            id_key = "relid"
            osm_id = osm_item.replace("relation/", "")
        elif osm_item.startswith("node"):
            queryName = "NodeGeo"
            id_key = "osm_id"
            osm_id = osm_item.replace("node/", "")
        else:
            raise ValueError(f"invalid osm_item {osm_item}")

        param_dict = {
            id_key: osm_id,
            "role": self.args.role,
            "min_lat": str(self.args.min_lat),
            "max_lat": str(self.args.max_lat),
            "min_lon": str(self.args.min_lon),
            "max_lon": str(self.args.max_lon),
        }

        if self.args.debug:
            print(f"Querying osm_item {osm_item}")

        lod = self.query_handler.query_by_name(
            query_name=queryName,
            param_dict=param_dict,
            endpoint=self.args.endpoint_name,
            auto_prefix=True,
        )
        return lod

    def compress_nodes(
        self, nodes: List[Dict], min_distance_m: float = 1000
    ) -> List[Dict]:
        """
        Compress nodes by removing ones
        that are too close together using haversine distance.

        Args:
            nodes: List of node data dicts
            min_distance_m: Minimum distance between nodes in meters

        Returns:
            List[Dict]: Filtered node data
        """
        if not nodes or len(nodes) <= 2:
            return nodes

        compressed_nodes = []
        last_lat, last_lon = None, None

        for node in nodes:
            wkt_node=node["wkt"]
            # get a copy
            wkt=dataclasses.replace(wkt_node)
            points = wkt.to_latlon_list()  # Extract all points from LINESTRING

            for lat, lon in points:
                if (
                    last_lat is None
                    or WKT.distance(last_lat, last_lon, lat, lon) >= min_distance_m
                ):
                    # keep reference
                    wkt.lat = lat
                    wkt.lon = lon
                    compressed_nodes.append(node)
                    last_lat, last_lon = lat, lon  # Update last kept point
                    break  # Move to the next node after selecting the first valid point

        return compressed_nodes

    def set_wkts(self, nodes: List[Dict]):
        """
        Convert loc entries to WKT instances for each node

        Args:
            nodes: List of node data dicts

        """
        for node in nodes:
            node["wkt"] = WKT(node["loc"])

    def to_mediawiki(self, osm_item: str, data: Dict) -> str:
        """
        Convert relation data to MediaWiki format

        Args:
             osm_item: Relation ID
             data: JSON data from query

        Returns:
             str: MediaWiki page content
        """
        wiki = f"""= Map =
https://www.openstreetmap.org/{osm_item}
{{{{LegMap
|zoom={self.args.zoom}
}}}}

= Locs =
"""
        # Add locations
        for item in data:
            node = item["node"]
            node_id = node.split("/")[-1]
            wkt = item["wkt"]
            latlon = wkt.to_latlon_str()
            name = item.get("node_name", node_id)
            loc = f"""{{{{Loc
|id={node_id}
|latlon={latlon}
|name={name}
|url={node}
|type={self.args.loc_type}
|address={self.args.country}
|storemode=subobject
}}}}
"""
            wiki += loc

        # Add legs
        wiki += "\n= Legs =\n"
        previous = None
        for item in data:
            if previous:
                leg = f"""{{{{Leg
|wp_num={item["rel_pos"]}
|from={previous["node"].split("/")[-1]}
|to={item["node"].split("/")[-1]}
|transport={self.args.transport}
|storemode=subobject
}}}}
"""
                wiki += leg
            previous = item

        wiki += f"\n<headertabs/>\n[[Category:{self.args.category}]]"
        return wiki

    def process_osm_items(self, osm_items: List[str], with_write: bool = True):
        """
        Process the given osm_items

        Args:
             osm_items: List of osm_items to process
             e.g. relation/<osm_id> or node/<osm_id>
        """
        for osm_item in osm_items:
            base_name = osm_item.replace("/", "_")
            json_file = os.path.join(self.tmpdir, f"osm_{base_name}.json")
            self.wiki_file = os.path.join(self.tmpdir, f"{base_name}.wiki")

            if not self.test:
                print(f"Processing osm item {osm_item}")

            # Query and save JSON
            q_data = self.query_osm_item(osm_item)
            if with_write:
                with open(json_file, "w") as f:
                    json.dump(q_data, f, indent=2)

            self.set_wkts(q_data)
            if self.args.min_node_distance:
                c_data = self.compress_nodes(
                    q_data, min_distance_m=self.args.min_node_distance
                )
            else:
                c_data = q_data

            # Convert to wiki and save
            wiki = self.to_mediawiki(osm_item, c_data)
            if with_write:
                with open(self.wiki_file, "w") as f:
                    f.write(wiki)

            if not self.test:
                print(f"Created {self.wiki_file}")
        return c_data


def main():
    """Main entry point"""
    args = Osm2WikiConverter.get_parser()
    converter = Osm2WikiConverter(args=args)
    converter.process_osm_items(args.osm_items)


if __name__ == "__main__":
    main()
